import os
from datetime import date, datetime

import minimalkv
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pyarrow as pa
import pytest
from pyarrow.parquet import ParquetFile

from plateau.core._compat import pandas_infer_string
from plateau.serialization import ParquetSerializer
from plateau.serialization._parquet import (
    MAX_NB_RETRIES,
    ParquetReadError,
    _predicate_accepts,
    _reset_dictionary_columns,
)
from plateau.serialization._util import _check_contains_null


@pytest.fixture
def reference_store():
    path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "..",
        "reference-data",
        "pyarrow-bugs",
    )
    return minimalkv.get_store_from_url(f"hfs://{path}")


@pytest.fixture(params=[1, 5, None])
def chunk_size(request, mocker):
    chunk_size = request.param
    if chunk_size == 1:
        # Test for a chunk size of one and mock the filter_df call. This way we can ensure that
        # the predicate for IO is properly evaluated and pushed down
        mocker.patch(
            "plateau.serialization._parquet.filter_df",
            new=lambda x, *args, **kwargs: x,
        )
    return chunk_size


@pytest.mark.parametrize("chunk_size", ["1", 1.0])
def test_parquet_serializer_raises_on_invalid_chunk_size_type(chunk_size):
    with pytest.raises(
        TypeError,
        match="Cannot initialize ParquetSerializer because chunk size is not integer type",
    ):
        ParquetSerializer(chunk_size=chunk_size)


@pytest.mark.parametrize("chunk_size", [-5, 0])
def test_parquet_serializer_raises_when_chunk_size_smaller_one(chunk_size):
    with pytest.raises(
        ValueError, match="Cannot initialize ParquetSerializer because chunk size < 1"
    ):
        ParquetSerializer(chunk_size=chunk_size)


@pytest.mark.parametrize("use_categorical", [True, False])
def test_rowgroup_writing(store, use_categorical, chunk_size):
    df = pd.DataFrame({"string": ["abc", "affe", "banane", "buchstabe"] * 5})
    serialiser = ParquetSerializer(chunk_size=chunk_size)
    # Arrow 0.9.0 has a bug in writing categorical columns to more than a single
    # RowGroup: "ArrowIOError: Column 2 had 2 while previous column had 4".
    # We have special handling for that in pandas-serialiser that should be
    # removed once we switch to 0.10.0
    if use_categorical:
        df_write = df.astype({"string": "category"})
    else:
        df_write = df
    key = serialiser.store(store, "prefix", df_write)

    parquet_file = ParquetFile(store.open(key))
    if chunk_size is None:
        assert parquet_file.num_row_groups == 1
    else:
        assert parquet_file.num_row_groups == 20 / chunk_size


_INT_TYPES = ["int8", "int16", "int32", "int64"]

# uint64 will fail since numexpr cannot safe case from uint64 to int64
_UINT_TYPES = ["uint8", "uint16", "uint32"]

_FLOAT_TYPES = ["float32", "float64"]

_STR_TYPES = ["unicode", "bytes"]

_DATE_TYPES = ["date"]
_DATETIME_TYPES = ["datetime64"]


def _validate_predicate_pushdown(df, column, value, store, chunk_size):
    serialiser = ParquetSerializer(chunk_size=chunk_size)
    key = serialiser.store(store, "prefix", df)

    predicates = [[(column, "==", value)]]

    df_restored = serialiser.restore_dataframe(store, key, predicates=predicates)
    # date objects are converted to datetime in pyarrow
    df_restored["date"] = df_restored["date"].dt.date

    expected = df.iloc[[3]]
    # ARROW-5138 index isn't preserved when doing predicate pushdown
    pdt.assert_frame_equal(
        df_restored.reset_index(drop=True), expected.reset_index(drop=True)
    )


@pytest.mark.parametrize("column", _INT_TYPES)
@pytest.mark.parametrize(
    "input_values",
    [
        (3, None),
        (3.0, TypeError),
        ("3", TypeError),
        ("3.0", TypeError),
        (b"3", TypeError),
        (b"3.0", TypeError),
    ],
)
def test_predicate_evaluation_integer(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


@pytest.mark.parametrize("column", _UINT_TYPES)
@pytest.mark.parametrize(
    "input_values",
    [
        (3, None),
        (3.0, TypeError),
        ("3", TypeError),
        ("3.0", TypeError),
        (b"3", TypeError),
        (b"3.0", TypeError),
    ],
)
def test_predicate_evaluation_unsigned_integer(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


@pytest.mark.parametrize("column", _FLOAT_TYPES)
@pytest.mark.parametrize(
    "input_values",
    [
        (3, TypeError),
        (3.0, None),
        ("3", TypeError),
        ("3.0", TypeError),
        (b"3", TypeError),
        (b"3.0", TypeError),
    ],
)
def test_predicate_evaluation_float(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


@pytest.mark.parametrize("column", _STR_TYPES)
@pytest.mark.parametrize(
    "input_values", [(3, TypeError), (3.0, TypeError), ("3", None), (b"3", None)]
)
def test_predicate_evaluation_string(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


@pytest.mark.parametrize("column", _DATE_TYPES)
@pytest.mark.parametrize(
    "input_values",
    [
        # it's the fifth due to the day % 31 in the testdata
        (date(2018, 1, 5), None),
        ("2018-01-05", None),
        (b"2018-01-05", None),
        (datetime(2018, 1, 1, 1, 1), TypeError),
        (3, TypeError),
        (3.0, TypeError),
        ("3", ValueError),
        ("3.0", ValueError),
        (b"3", ValueError),
        (b"3.0", ValueError),
    ],
)
def test_predicate_evaluation_date(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


@pytest.mark.parametrize("column", _DATETIME_TYPES)
@pytest.mark.parametrize(
    "input_values",
    [
        (datetime(2018, 1, 5), None),
        (np.datetime64(datetime(2018, 1, 5)), None),
        (pd.Timestamp(datetime(2018, 1, 5)), None),
        (np.datetime64(datetime(2018, 1, 5), "s"), None),
        (np.datetime64(datetime(2018, 1, 5), "ms"), None),
        (np.datetime64(datetime(2018, 1, 5), "us"), None),
        (np.datetime64(datetime(2018, 1, 5), "ns"), None),
        (date(2018, 1, 4), TypeError),
        ("2018-01-04", TypeError),
        (b"2018-01-04", TypeError),
        (1, TypeError),
        (1.0, TypeError),
    ],
)
def test_predicate_evaluation_datetime(
    store, dataframe_not_nested, column, input_values, chunk_size
):
    value, exception = input_values
    if exception:
        with pytest.raises(exception):
            _validate_predicate_pushdown(
                dataframe_not_nested, column, value, store, chunk_size
            )
    else:
        _validate_predicate_pushdown(
            dataframe_not_nested, column, value, store, chunk_size
        )


def test_ensure_binaries(binary_value):
    assert isinstance(binary_value, bytes)


def test_pushdown_binaries(store, dataframe_not_nested, binary_value, chunk_size):
    if _check_contains_null(binary_value):
        pytest.xfail("Null-terminated binary strings are not supported")
    serialiser = ParquetSerializer(chunk_size=chunk_size)
    key = serialiser.store(store, "prefix", dataframe_not_nested)

    predicates = [[("bytes", "==", binary_value)]]

    df_restored = serialiser.restore_dataframe(store, key, predicates=predicates)
    assert len(df_restored) == 1
    assert df_restored.iloc[0].bytes == binary_value


@pytest.mark.xfail(reason="Requires parquet-cpp 1.5.0.")
def test_pushdown_null_intermediate(store):
    binary = b"\x8f\xb6\xe5@\x90\xdc\x11\xe8\x00\xae\x02B\xac\x12\x01\x06"
    df = pd.DataFrame({"byte_with_null": [binary]})
    serialiser = ParquetSerializer(chunk_size=1)
    key = serialiser.store(store, "key", df)
    predicate = [[("byte_with_null", "==", binary)]]
    restored = serialiser.restore_dataframe(store, key, predicates=predicate)
    pdt.assert_frame_equal(restored, df)


@pytest.mark.parametrize("chunk_size", [None, 1])
def test_date_as_object(store, chunk_size):
    ser = ParquetSerializer(chunk_size=chunk_size)
    df = pd.DataFrame({"date": [date(2000, 1, 1), date(2000, 1, 2)]})
    key = ser.store(store, "key", df)
    restored_df = ser.restore_dataframe(
        store, key, categories=["date"], date_as_object=True
    )
    categories = pd.Series([date(2000, 1, 1), date(2000, 1, 2)])
    expected_df = pd.DataFrame({"date": pd.Categorical(categories)})
    # expected_df.date = expected_df.date.cat.rename_categories([date(2000, 1, 1)])
    pdt.assert_frame_equal(restored_df, expected_df)

    restored_df = ser.restore_dataframe(
        store, key, date_as_object=True, predicates=[[("date", "==", "2000-01-01")]]
    )
    expected_df = pd.DataFrame({"date": [date(2000, 1, 1)]})
    pdt.assert_frame_equal(restored_df, expected_df)


@pytest.mark.parametrize("chunk_size", [None, 1])
def test_predicate_not_in_columns(store, chunk_size):
    ser = ParquetSerializer(chunk_size=chunk_size)
    df = pd.DataFrame(
        {
            "date": [date(2000, 1, 1), date(2000, 1, 2), date(2000, 1, 2)],
            "col": [1, 2, 1],
        }
    )
    key = ser.store(store, "key", df)
    restored_df = ser.restore_dataframe(
        store, key, columns=[], predicates=[[("col", "==", 1)]]
    )
    if chunk_size:
        expected_df = pd.DataFrame(
            index=[0, 1],
            columns=pd.Index([], dtype=str if pandas_infer_string() else "O"),
        )
    else:
        expected_df = pd.DataFrame(
            index=[0, 2],
            columns=pd.Index([], dtype=str if pandas_infer_string() else "O"),
        )

    pdt.assert_frame_equal(restored_df, expected_df)


def test_read_empty_file_with_predicates(store):
    ser = ParquetSerializer()
    df = pd.DataFrame({"col": pd.Series([], dtype=str)})
    key = ser.store(store, "key", df)
    restored_df = ser.restore_dataframe(
        store, key, columns=["col"], predicates=[[("col", "==", "1")]]
    )
    pdt.assert_frame_equal(restored_df, df)


@pytest.mark.parametrize("predicate_pushdown_to_io", [True, False])
def test_int64_statistics_overflow(reference_store, predicate_pushdown_to_io):
    # Test case for ARROW-5166
    ser = ParquetSerializer()

    v = 705449463447499237
    predicates = [[("x", "==", v)]]
    result = ser.restore_dataframe(
        reference_store,
        "int64_statistics_overflow.parquet",
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        predicates=predicates,
    )
    assert not result.empty
    assert (result["x"] == v).all()


@pytest.mark.parametrize(
    ["predicate_value", "expected"],
    [
        ([0, 4, 1], True),
        ([-2, 44], False),
        ([-3, 0], True),
        ([-1, 10**4], False),
        ([2, 3], True),
        ([-1, 20], True),
        ([-30, -5, 50, 10], True),
        ([], False),
    ],
)
def test_predicate_accept_in(store, predicate_value, expected):
    df = pd.DataFrame({"A": [0, 4, 13, 29]})  # min = 0, max = 29
    predicate = ("A", "in", predicate_value)
    serialiser = ParquetSerializer(chunk_size=None)
    key = serialiser.store(store, "prefix", df)

    parquet_file = ParquetFile(store.open(key))
    row_meta = parquet_file.metadata.row_group(0)
    arrow_schema = parquet_file.schema.to_arrow_schema()
    parquet_reader = parquet_file.reader
    assert (
        _predicate_accepts(
            predicate,
            row_meta=row_meta,
            arrow_schema=arrow_schema,
            parquet_reader=parquet_reader,
        )
        == expected
    )


def test_read_categorical(store):
    df = pd.DataFrame({"col": ["a"]}).astype({"col": "category"})

    serialiser = ParquetSerializer()
    key = serialiser.store(store, "prefix", df)

    df = serialiser.restore_dataframe(store, key)
    assert df.dtypes["col"] == "O" if not pandas_infer_string() else "string"

    df = serialiser.restore_dataframe(store, key, categories=["col"])
    assert df.dtypes["col"] == pd.CategoricalDtype(["a"], ordered=False)


def test_read_empty_categorical(store):
    df = pd.DataFrame({"col": [None]}).astype({"col": "category"})

    serialiser = ParquetSerializer()
    key = serialiser.store(store, "prefix", df)

    df = serialiser.restore_dataframe(store, key)
    assert df.dtypes["col"] == "O" if not pandas_infer_string() else "string"

    df = serialiser.restore_dataframe(store, key, categories=["col"])
    assert df.dtypes["col"] == pd.CategoricalDtype([], ordered=False)


def test_read_categorical_empty_dataframe(store):
    df = pd.DataFrame({"col": ["a"]}).astype({"col": "category"}).iloc[:0]
    serialiser = ParquetSerializer()
    key = serialiser.store(store, "prefix", df)

    df = serialiser.restore_dataframe(store, key)
    assert df.dtypes["col"] == "O" if not pandas_infer_string() else "string"

    df = serialiser.restore_dataframe(store, key, categories=["col"])

    assert df.dtypes["col"] == pd.CategoricalDtype([], ordered=False)


def test_reset_dict_cols(store):
    df = pd.DataFrame({"col": ["a"], "colB": ["b"]}).astype(
        {"col": "category", "colB": "category"}
    )
    table = pa.Table.from_pandas(df)
    orig_schema = table.schema

    assert pa.types.is_dictionary(orig_schema.field("col").type)
    assert pa.types.is_dictionary(orig_schema.field("colB").type)

    all_reset = _reset_dictionary_columns(table).schema
    assert not pa.types.is_dictionary(all_reset.field("col").type)
    assert not pa.types.is_dictionary(all_reset.field("colB").type)

    only_a_reset = _reset_dictionary_columns(table, exclude=["colB"]).schema
    assert not pa.types.is_dictionary(only_a_reset.field("col").type)
    assert pa.types.is_dictionary(only_a_reset.field("colB").type)


def test_retry_on_IOError(monkeypatch, caplog, store):  # noqa: N802
    """See https://github.com/JDASoftwareGroup/plateau/issues/407 :

    We are testing a retry-workaround for the above issue here. Once the
    issue is resolved, this test and the workaround can be removed.
    """

    df = pd.DataFrame({"A": [0, 1, 2, 3]})

    retry_count = 0

    def patched__restore_dataframe(**kwargs):
        nonlocal retry_count
        retry_count += 1

        if not retry_count > 1:
            # fail for the first try
            raise OSError()
        elif retry_count > 1:
            # simulate a successful retry
            return df

    monkeypatch.setattr(
        ParquetSerializer, "_restore_dataframe", patched__restore_dataframe
    )
    serializer = ParquetSerializer()
    key = serializer.store(store, "key", df)
    df_result = serializer.restore_dataframe(store=store, key=key)
    pdt.assert_frame_equal(df, df_result)


def test_retries_on_IOError_logs(monkeypatch, caplog, store):  # noqa: N802
    """See https://github.com/JDASoftwareGroup/plateau/issues/407 :

    We are testing a retry-workaround for the above issue here. Once the
    issue is resolved, this test and the workaround can be removed.
    """

    def patched__restore_dataframe(**kwargs):
        # This kind of exception should be captured by the retry mechanism.
        raise OSError()

    df = pd.DataFrame({"A": [0, 1, 2, 3]})
    monkeypatch.setattr(
        ParquetSerializer, "_restore_dataframe", patched__restore_dataframe
    )
    serializer = ParquetSerializer()
    key = serializer.store(store, "key", df)

    with pytest.raises(ParquetReadError):
        serializer.restore_dataframe(store=store, key=key)

    assert len(caplog.records) == MAX_NB_RETRIES
    for log_record in caplog.records:
        assert "Failed to restore dataframe" in log_record.message


def test_retry_fail_on_other_error(monkeypatch, caplog, store):
    """See https://github.com/JDASoftwareGroup/plateau/issues/407 : We are
    testing a retry-workaround for the above issue here. Once the issue is
    resolved, this test and the workaround can be removed.

    We only want to retry on OSErrors (and inherited exceptions) -- all
    other exceptions should be raised.
    """

    df = pd.DataFrame({"A": [0, 1, 2, 3]})

    def patched__restore_dataframe(**kwargs):
        # This should not be retried but raised immediately.
        raise ValueError()

    monkeypatch.setattr(
        ParquetSerializer, "_restore_dataframe", patched__restore_dataframe
    )
    serializer = ParquetSerializer()
    key = serializer.store(store, "key", df)

    with pytest.raises(ValueError):
        serializer.restore_dataframe(store=store, key=key)

    assert len(caplog.records) == 0



import datetime

import numpy as np
import pandas as pd
import pytest
from pandas import testing as pdt

from plateau.serialization import filter_array_like, filter_df_from_predicates


@pytest.fixture(
    params=[
        np.array(range(10)),
        np.array(range(10)).astype(float),
        pd.Series(np.array(range(10))),
        pd.Series(np.array(range(10)).astype(float)),
        pd.Series(np.array([str(x) for x in range(10)])),
        pd.Series(np.array([str(x) for x in range(10)])).astype("category"),
        np.array(["A", "B", "C", "D", "E", "F"], dtype="<U1"),
        np.array(["A", "B", "C", "D", "E", "F"], dtype="S"),
    ]
)
def array_like(request):
    return request.param


@pytest.fixture
def df_none():
    return pd.DataFrame({"A": [True, False, True, True, None, False] * 2})


@pytest.fixture
def df_na():
    return pd.DataFrame(
        {"A": [0.0, 0.0, np.nan, np.nan, 1.0, 1.0, 1.0, 1.0, np.nan, 2.0, 2.0]}
    )


def test_filter_array_like_eq(array_like):
    ix = 3
    value = array_like[ix]
    res = filter_array_like(array_like, "==", value)

    assert all(array_like[res] == array_like[ix])


def test_filter_array_like_larger_eq(array_like):
    ix = 3
    value = array_like[ix]
    res = filter_array_like(array_like, ">=", value)

    assert all(array_like[res] == array_like[ix:])


def test_filter_array_like_lower_eq(array_like):
    ix = 3
    value = array_like[ix]
    res = filter_array_like(array_like, "<=", value)

    assert all(array_like[res] == array_like[: ix + 1])


@pytest.mark.parametrize(
    "op, expected",
    [
        (">", [False, True, False]),
        (">=", [True, True, False]),
        ("<=", [True, False, True]),
        ("<", [False, False, True]),
    ],
)
@pytest.mark.parametrize(
    "cat_type",
    [
        "category",
        pd.CategoricalDtype(["A", "B", "C"], ordered=False),
        pd.CategoricalDtype(["B", "C", "A"], ordered=False),
        pd.CategoricalDtype(["A", "B", "C"], ordered=True),
    ],
)
def test_filter_array_like_categoricals(op, expected, cat_type):
    ser = pd.Series(["B", "C", "A"]).astype(cat_type)
    res = filter_array_like(ser, op, "B")

    assert all(res == expected)


@pytest.mark.parametrize(
    "value, filter_value",
    [
        (1, 1.0),
        ("1", 1.0),
        ("1", 1),
        (1, "1"),
        (datetime.date(2019, 1, 1), 1),
        (datetime.datetime(2019, 1, 1), 1),
        (datetime.datetime(2019, 1, 1), datetime.date(2019, 1, 1)),
        (datetime.date(2019, 1, 1), datetime.datetime(2019, 1, 1)),
        (True, 1),
        (True, 1.0),
        (True, "True"),
        (True, None),
        (datetime.datetime(2019, 1, 1), True),
        ("2019-01-01", datetime.datetime(2019, 1, 1)),
        # we are allowing object arrays comparison with a boolean value
        pytest.param("True", True, marks=pytest.mark.xfail(reason="see gh-193")),
        pytest.param(b"True", True, marks=pytest.mark.xfail(reason="see gh-193")),
        pytest.param([True], True, marks=pytest.mark.xfail(reason="see gh-193")),
    ],
)
@pytest.mark.parametrize("op", ["==", "!=", "<", "<=", ">", ">=", "in"])
def test_raise_on_type(value, filter_value, op):
    array_like = pd.Series([value])
    with pytest.raises(TypeError, match="Unexpected type for predicate:"):
        filter_array_like(array_like, op, filter_value, strict_date_types=True)


@pytest.mark.parametrize("op", ["==", "!=", ">=", "<=", ">", "<", "in"])
@pytest.mark.parametrize(
    "data,value",
    [
        (
            # data
            range(10),
            # value
            4,
        ),
        (
            # data
            ["A", "B"] * 5,
            # value
            "A",
        ),
        (
            # data
            pd.Series(["X", "Y"] * 5).astype("category"),
            # value
            "X",
        ),
        (
            # data
            pd.Series([datetime.date(2019, 1, 1), datetime.date(2019, 1, 2)] * 5),
            # value
            datetime.date(2019, 1, 1),
        ),
        (
            # data
            [datetime.datetime(2019, 1, 1), datetime.datetime(2019, 1, 2)] * 5,
            # value
            datetime.datetime(2019, 1, 1),
        ),
        (
            # data
            [datetime.datetime(2019, 1, 1), datetime.datetime(2019, 1, 2)] * 5,
            # value
            datetime.date(2019, 1, 1),
        ),
        (
            # data
            np.arange(10, dtype=np.uint8),
            # value
            4,
        ),
    ],
)
def test_filter_df_from_predicates(op, data, value):
    df = pd.DataFrame({"A": data})
    df["B"] = range(len(df))

    if isinstance(df["A"].dtype, pd.CategoricalDtype):
        df["A"] = df["A"].astype(df["A"].cat.as_ordered().dtype)

    if op == "in":
        value = [value]

    predicates = [[("A", op, value)]]
    actual = filter_df_from_predicates(df, predicates)
    # Note that the predicates above include the value without any casting.
    # plateau is actually more forgiving w.r.t. matching data types. For the
    # eval below we want to ensure that the types match exactly
    value = pd.Series(value, dtype=df["A"].dtype).iloc[0]
    if op == "in":
        expected = df[df["A"].isin([value])]
    else:
        expected = eval(f"df[df['A'] {op} value]")
    pdt.assert_frame_equal(actual, expected, check_categorical=False)


@pytest.mark.parametrize("op", ["==", "!=", "<", "<=", ">", ">="])
@pytest.mark.parametrize("unit_input", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("unit_value", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize(
    "type_value",
    [str, pd.Timestamp, datetime.datetime, np.datetime64],
)
def test_filter_datetime_varying_units(unit_input, unit_value, op, type_value):
    datetime_series = pd.Series(
        pd.date_range("2019-01-01", periods=10, freq="D", unit=unit_input)
    )

    value = pd.Timestamp(datetime_series[5], unit=unit_value)
    if type_value is str:
        value = value.isoformat()
    elif type_value is datetime.datetime:
        value = value.to_pydatetime()
    elif type_value is np.datetime64:
        value = value.to_datetime64()
    elif type_value is pd.Timestamp:
        # pd.Timestamp already in the correct format
        pass

    data_ns = datetime_series.astype("datetime64[ns]")
    expected = filter_array_like(data_ns, op, data_ns[5])
    actual = filter_array_like(datetime_series, op, value)

    assert (actual == expected).all()


@pytest.mark.parametrize("op", ["==", "!="])
@pytest.mark.parametrize("col", list("AB"))
def test_filter_df_from_predicates_bool(op, col):
    df = pd.DataFrame(
        {"A": [True, False] * 5, "B": [True, False, None, True, False] * 2}
    )

    value = True
    predicates = [[(col, op, value)]]
    actual = filter_df_from_predicates(df, predicates)
    if isinstance(df[col].dtype, pd.CategoricalDtype):
        df[col] = df[col].astype(df[col].cat.as_ordered().dtype)
    expected = eval(f"df[df[col] {op} value]")
    pdt.assert_frame_equal(actual, expected, check_categorical=False)


@pytest.mark.parametrize(
    "op, value, expected",
    [
        # != will be the same as `is distinct from` until 5.0.0.
        ("==", True, "df_none[df_none['A'] == True]"),
        ("!=", True, "df_none[df_none['A'] != True]"),
        ("is distinct from", True, "df_none[df_none['A'] != True]"),
        ("==", False, "df_none[df_none['A'] == False]"),
        ("!=", False, "df_none[df_none['A'] != False]"),
        ("is distinct from", False, "df_none[df_none['A'] != False]"),
        ("==", None, "df_none[df_none['A'].isnull()]"),
        ("!=", None, "df_none[df_none['A'].notnull()]"),
        ("is distinct from", None, "df_none[df_none['A'].notnull()]"),
    ],
)
def test_filter_df_from_predicates_missing_bool(df_none, op, value, expected):
    predicates = [[("A", op, value)]]
    actual = filter_df_from_predicates(df_none, predicates)
    pdt.assert_frame_equal(actual, eval(expected))


@pytest.mark.parametrize(
    "op, value, expected",
    [
        # != will be the same as `is distinct from` until 5.0.0.
        ("==", 0.0, "df_na[df_na['A'] == 0.0]"),
        ("!=", 0.0, "df_na[df_na['A'] != 0.0]"),
        ("is distinct from", 0.0, "df_na[df_na['A'] != 0.0]"),
        ("==", 1.0, "df_na[df_na['A'] == 1.0]"),
        ("!=", 1.0, "df_na[df_na['A'] != 1.0]"),
        ("is distinct from", 1.0, "df_na[df_na['A'] != 1.0]"),
        ("==", np.nan, "df_na[df_na['A'].isnull()]"),
        ("!=", np.nan, "df_na[df_na['A'].notnull()]"),
        ("is distinct from", np.nan, "df_na[df_na['A'].notnull()]"),
    ],
)
def test_filter_df_from_predicates_missing_scalar(df_na, op, value, expected):
    predicates = [[("A", op, value)]]
    actual = filter_df_from_predicates(df_na, predicates)
    pdt.assert_frame_equal(actual, eval(expected))


def test_not_equal_predicate_gives_deprecation_warning(df_na):
    predicates = [[("A", "!=", 1.0)]]
    msg = (
        "The current behaviour of the != will be deprecated in version 5.0.0. "
        "The new behaviour will be SQL-like, i.e. NaN values will be filtered out."
    )
    with pytest.deprecated_call(match=msg):
        actual = filter_df_from_predicates(df_na, predicates)
    pdt.assert_frame_equal(actual, df_na[df_na["A"] != 1.0])


@pytest.mark.parametrize(
    "value",
    [
        1,
        np.uint8(1),
        1.1,
        "A",
        datetime.date(2020, 1, 1),
        pd.Timestamp("2020-01-01"),
        np.datetime64("2020-01-01"),
    ],
)
def test_filter_df_from_predicates_empty_in(value):
    df = pd.DataFrame({"A": [value]})
    df["B"] = range(len(df))

    predicates = [[("A", "in", [])]]
    actual = filter_df_from_predicates(df, predicates)
    expected = df.iloc[[]]
    pdt.assert_frame_equal(actual, expected, check_categorical=False)


def test_filter_df_from_predicates_or_predicates():
    df = pd.DataFrame({"A": range(10), "B": ["A", "B"] * 5, "C": range(-10, 0)})

    predicates = [[("A", "<", 3)], [("A", ">", 5)], [("B", "==", "non-existent")]]
    actual = filter_df_from_predicates(df, predicates)
    expected = pd.DataFrame(
        data={
            "A": [0, 1, 2, 6, 7, 8, 9],
            "B": ["A", "B", "A", "A", "B", "A", "B"],
            "C": [-10, -9, -8, -4, -3, -2, -1],
        },
        index=[0, 1, 2, 6, 7, 8, 9],
    )
    pdt.assert_frame_equal(actual, expected)

    predicates = [[("A", "<", 3)], [("A", ">", 5)], [("B", "==", "B")]]
    actual = filter_df_from_predicates(df, predicates)
    # row for (A == 4) is filtered out
    expected = pd.DataFrame(
        data={
            "A": [0, 1, 2, 3, 5, 6, 7, 8, 9],
            "B": ["A", "B", "A", "B", "B", "A", "B", "A", "B"],
            "C": [-10, -9, -8, -7, -5, -4, -3, -2, -1],
        },
        index=[0, 1, 2, 3, 5, 6, 7, 8, 9],
    )
    pdt.assert_frame_equal(actual, expected)



import datetime

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
import simplejson

import plateau.core._zmsgpack as msgpack
from plateau.core.common_metadata import make_meta, store_schema_metadata
from plateau.core.dataset import DatasetMetadata
from plateau.core.index import ExplicitSecondaryIndex, PartitionIndex
from plateau.core.testing import cm_frozen_time

# Basic functionality tests.
#
# In test_reference_data.py we test even more constellations of datasets.

TIME_TO_FREEZE = datetime.datetime(2000, 1, 1, 1, 1, 1, 1)


@pytest.fixture
def frozen_time():
    with cm_frozen_time(TIME_TO_FREEZE):
        yield


def test_roundtrip_empty(metadata_version):
    ds = DatasetMetadata(uuid="dataset_uuid", metadata_version=metadata_version)
    assert ds == ds.from_dict(ds.to_dict())


def test_roundtrip_empty_with_store(store, metadata_version):
    dataset_uuid = "dataset_uuid"
    dataset = DatasetMetadata(uuid=dataset_uuid, metadata_version=metadata_version)
    store.put(
        f"{dataset_uuid}.by-dataset-metadata.json",
        simplejson.dumps(dataset.to_dict()).encode("utf-8"),
    )
    assert dataset == DatasetMetadata.load_from_store(dataset_uuid, store)


def test_roundtrip(metadata_version, frozen_time):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid",
        "metadata": {"key": "value", "creation_time": "2000-01-01 01:01:01"},
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
        "partition_keys": [],
        "indices": {"p_id": {"1": ["part_1"]}},
    }
    result = DatasetMetadata.from_dict(expected).to_dict()
    assert expected == result


def test_roundtrip_no_metadata(metadata_version, frozen_time):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid",
        "metadata": {"creation_time": "2000-01-01 01:01:01"},
        "partition_keys": [],
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
    }
    result = DatasetMetadata.from_dict(expected).to_dict()
    assert expected == result


def test_roundtrip_json(metadata_version):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid",
        "metadata": {"key": "value", "creation_time": "2000-01-01 01:01:01"},
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
        "partition_keys": [],
        "indices": {"p_id": {"1": ["part_1"]}},
    }

    result = simplejson.loads(
        DatasetMetadata.from_buffer(simplejson.dumps(expected)).to_json()
    )
    assert expected == result


def test_raise_multitable(metadata_version):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid",
        "metadata": {},
        "partitions": {
            "part_1": {"files": {"tableA": "file.parquet", "tableB": "file.parquet"}}
        },
        "partition_keys": [],
    }

    with pytest.raises(
        RuntimeError,
        match=r"Dataset uuid has tables.*but read support for multi tabled dataset was dropped with plateau 4\.0\.",
    ):
        DatasetMetadata.from_dict(expected)


def test_roundtrip_msgpack():
    expected = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "uuid",
        "metadata": {"key": "value", "creation_time": "2000-01-01 01:01:01"},
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
        "partition_keys": [],
        "indices": {"p_id": {"1": ["part_1"]}},
    }

    result = msgpack.unpackb(
        DatasetMetadata.from_buffer(
            msgpack.packb(expected), format="msgpack"
        ).to_msgpack()
    )
    assert expected == result


def test_invalid_uuid():
    expected = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "uuid.",
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
    }
    with pytest.raises(ValueError):
        DatasetMetadata.from_dict(expected)

    expected = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "mañana",
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
    }
    with pytest.raises(ValueError):
        DatasetMetadata.from_dict(expected)


def test_complicated_uuid():
    expected = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "uuid+namespace-attribute12_underscored",
        "partitions": {"part_1": {"files": {"core": "file.parquet"}}},
    }
    DatasetMetadata.from_dict(expected)


def test_read_table_meta(store):
    meta_dct = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "dataset_uuid",
        "partitions": {
            "location_id=1/part_1": {
                "files": {"table1": "dataset_uuid/table1/location_id=1/part_1.parquet"}
            }
        },
    }
    df1 = pd.DataFrame(
        {"location_id": pd.Series([1], dtype=int), "x": pd.Series([True], dtype=bool)}
    )
    schema1 = make_meta(df1, origin="1")
    store_schema_metadata(schema1, "dataset_uuid", store, "table1")

    dmd = DatasetMetadata.load_from_dict(meta_dct, store)

    assert dmd.schema == schema1


def test_load_indices_embedded(metadata_version):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid+namespace-attribute12_underscored",
        "partitions": {"part_1": {"files": {"core_data": "file.parquest"}}},
        "indices": {
            "product_id": {
                "1": ["part_1"],
                "2": ["part_1"],
                "100": ["part_1"],
                "34": ["part_1"],
            }
        },
    }
    dmd = DatasetMetadata.from_dict(expected)
    assert "product_id" in dmd.indices

    with pytest.raises(KeyError):
        dmd.load_index("not there", store=None)

    dmd_loaded = dmd.load_index("product_id", store=None)
    assert "product_id" in dmd_loaded.indices


def test_load_all_indices(store, metadata_version):
    meta_dct = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid+namespace-attribute12_underscored",
        "partitions": {
            "location_id=1/part_1": {
                "files": {
                    "core_data": "dataset_uuid/table/location_id=1/part_1.parquet"
                }
            }
        },
        "indices": {
            "product_id": {
                "1": ["part_1"],
                "2": ["part_1"],
                "100": ["part_1"],
                "34": ["part_1"],
            }
        },
    }
    dmd = DatasetMetadata.from_dict(meta_dct)
    dmd.schema = make_meta(
        pd.DataFrame({"location_id": pd.Series([1], dtype=int)}), origin="core"
    )

    dmd = dmd.load_all_indices(store)

    assert "product_id" in dmd.indices
    assert isinstance(dmd.indices["product_id"], ExplicitSecondaryIndex)

    assert "location_id" in dmd.indices
    assert isinstance(dmd.indices["location_id"], PartitionIndex)

    assert len(dmd.indices) == 2


def test_load_from_store_with_indices(store):
    meta_dct = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "uuid",
        "partitions": {
            "product_id=1/part_1": {
                "files": {
                    "core_data": "dataset_uuid/table/location_id=1/part_1.parquet"
                }
            }
        },
        "indices": {
            "product_id": {
                "1": ["part_1"],
                "2": ["part_1"],
                "100": ["part_1"],
                "34": ["part_1"],
            }
        },
    }
    store.put(
        "uuid.by-dataset-metadata.json", simplejson.dumps(meta_dct).encode("utf-8")
    )
    df = pd.DataFrame({"index": [1], "location_id": [1], "product_id": [1]})
    store_schema_metadata(make_meta(df, origin="core"), "uuid", store, "core_data")

    storage_key = "uuid/some_index.parquet"
    index2 = ExplicitSecondaryIndex(
        column="location_id",
        index_dct={1: ["part_1", "part_2"], 3: ["part_3"]},
        index_storage_key=storage_key,
        dtype=pa.int64(),
    )
    index2.store(store, "dataset_uuid")

    dmd = DatasetMetadata.load_from_store(store=store, uuid="uuid")
    assert "location_id" not in dmd.indices

    dmd = DatasetMetadata.load_from_store(
        store=store, uuid="uuid", load_all_indices=True
    )
    assert "location_id" in dmd.indices


def test_load_partition_indices_types(store):
    dataset_uuid = "uuid+namespace-attribute12_underscored"
    table = "table"
    index_name = "location_id"
    index_value = 1
    meta_dct = {
        "dataset_metadata_version": 4,
        "dataset_uuid": dataset_uuid,
        "partitions": {
            f"{index_name}={index_value}/part_1": {
                "files": {table: f"{dataset_uuid}/{table}/location_id=1/part_1.parquet"}
            }
        },
    }
    store.put(
        f"{dataset_uuid}.by-dataset-metadata.json",
        simplejson.dumps(meta_dct).encode(),
    )
    store_schema_metadata(
        make_meta(
            pd.DataFrame({index_name: pd.Series([index_value], dtype=int)}),
            origin="core",
        ),
        dataset_uuid,
        store,
        table,
    )
    dmd = DatasetMetadata.load_from_store(store=store, uuid=dataset_uuid)

    dmd = dmd.load_partition_indices()
    assert len(dmd.indices) == 1

    assert "location_id" in dmd.indices
    assert isinstance(dmd.indices["location_id"], PartitionIndex)

    idx = dmd.indices["location_id"]
    assert idx.dtype == pa.int64()
    assert idx.query(1) == ["location_id=1/part_1"]


def test_load_partition_keys(store):
    expected = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "uuid",
        "partitions": {
            "part_1": {
                "files": {"core_data": "uuid/table/index=1/index2=2/file.parquet"}
            },
            "part_2": {
                "files": {"core_data": "uuid/table/index=1/index2=2/file2.parquet"}
            },
        },
        "indices": {
            "product_id": {
                "1": ["part_1"],
                "2": ["part_2"],
                "100": ["part_1", "part_2"],
                "34": ["part_1"],
            },
            "location_id": {
                "1": ["part_1"],
                "2": ["part_2"],
                "3": ["part_1"],
                "4": ["part_2"],
            },
        },
    }
    store.put(
        "uuid.by-dataset-metadata.json", simplejson.dumps(expected).encode("utf-8")
    )
    df = pd.DataFrame(
        {"index": [1], "index2": [1], "product_id": [1], "location_id": [1]}
    )
    store_schema_metadata(make_meta(df, origin="core"), "uuid", store, "core_data")
    dmd = DatasetMetadata.load_from_store("uuid", store)
    assert dmd.partition_keys == ["index", "index2"]


def test_query_indices_external(store, metadata_version):
    expected = {
        "dataset_metadata_version": metadata_version,
        "dataset_uuid": "uuid+namespace-attribute12_underscored",
        "partitions": {
            "part_1": {"files": {"core_data": "file.parquest"}},
            "part_2": {"files": {"core_data": "file2.parquest"}},
        },
        "indices": {
            "product_id": "uuid+namespace-attribute12_underscored.product_id.by-dataset-index.parquet",
            "location_id": {
                "1": ["part_1"],
                "2": ["part_2"],
                "3": ["part_1"],
                "4": ["part_2"],
            },
        },
    }
    store.put(
        "uuid+namespace-attribute12_underscored.by-dataset-metadata.json",
        simplejson.dumps(expected).encode("utf-8"),
    )
    df = pd.DataFrame(
        {
            "product_id": [1, 2, 100, 34],
            "partition": [
                np.array(["part_1"], dtype=object),
                np.array(["part_2"], dtype=object),
                np.array(["part_1", "part_2"], dtype=object),
                np.array(["part_1"], dtype=object),
            ],
        }
    )
    schema = pa.schema(
        [
            pa.field("partition", pa.list_(pa.string())),
            pa.field("product_id", pa.int64()),
        ]
    )
    table = pa.Table.from_pandas(df, schema=schema)
    buf = pa.BufferOutputStream()
    pq.write_table(table, buf)
    store.put(
        "uuid+namespace-attribute12_underscored.product_id.by-dataset-index.parquet",
        buf.getvalue().to_pybytes(),
    )
    store_schema_metadata(
        make_meta(df, origin="core"),
        "uuid+namespace-attribute12_underscored",
        store,
        "core_data",
    )

    dmd = DatasetMetadata.load_from_store(
        "uuid+namespace-attribute12_underscored", store
    )

    dmd = dmd.load_index("product_id", store)
    assert dmd.query(product_id=2) == ["part_2"]
    dmd = dmd.load_all_indices(store)
    assert dmd.query(product_id=2, location_id=2) == ["part_2"]
    assert dmd.query(product_id=100, location_id=3) == ["part_1"]
    assert dmd.query(product_id=2, location_id=2, something_else="bla") == ["part_2"]

    additional_index = ExplicitSecondaryIndex.from_v2(
        "another_column", {"1": ["part_2", "part_3"]}
    )
    assert dmd.query(
        indices=[additional_index], another_column="1", product_id=2, location_id=2
    ) == ["part_2"]


def test_copy(frozen_time):
    ds = DatasetMetadata(
        uuid="uuid",
        partitions={"partition_label": {"files": {}}},
        metadata={"some": "metadata"},
        indices={
            "column": ExplicitSecondaryIndex(
                column="column", index_dct={1: ["partition_label"]}
            )
        },
        explicit_partitions=True,
        partition_keys=["P", "L"],
    )
    new_ds = ds.copy()
    # Check if the copy is identical
    assert new_ds == ds
    # ... but not the same object
    assert id(new_ds) != id(ds)

    new_ds = ds.copy(metadata={"new": "metadata"})
    assert id(new_ds) != id(ds)
    assert new_ds.metadata == {
        "new": "metadata",
        # The DatasetMetadata constructor ensure that the creation time is
        # always present.
        "creation_time": "2000-01-01T01:01:01.000001",
    }


def test_load_partition_indices_no_files(store):
    meta_dct = {
        "dataset_metadata_version": 4,
        "dataset_uuid": "dataset_uuid",
        "partitions": {"p1": {"files": {}}},
    }
    dmd = DatasetMetadata.load_from_dict(meta_dct, store)
    dmd = dmd.load_partition_indices()
    assert len(dmd.indices) == 0


def test_load_missing(store):
    with pytest.raises(KeyError) as exc:
        DatasetMetadata.load_from_store("not_there", store)
    assert exc.value.args == (
        "Dataset does not exist. Tried not_there.by-dataset-metadata.json and not_there.by-dataset-metadata.msgpack.zstd",
    )



def test_docs_use_api():
    import os

    files_to_check = []
    for root, _, files in os.walk("../docs"):
        for f in files:
            if f.endswith(".rst"):
                files_to_check.append(os.path.join(root, f))

    for file_ in files_to_check:
        with open(file_) as fd:
            content = fd.read()

        pattern = r"(from|import) plateau\.(?!(api))"
        import re

        if re.search(pattern, content):
            raise AssertionError(f"Found non-api import in document {file_}")



#!/usr/bin/env python
"""This module contains functionality for persisting/serialising DataFrames.

Available constants

**PredicatesType** - A type describing the format of predicates which is a list of ConjunctionType
**ConjunctionType** - A type describing a single Conjunction which is a list of literals
**LiteralType**  - A type for a single literal

**LiteralValue** - A type indicating the value of a predicate literal


:meta public:
"""

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING, TypeVar

import numpy as np
import pandas as pd
from minimalkv import KeyValueStore
from pandas.api.types import is_list_like

from plateau.serialization._util import _check_contains_null

from ._util import ensure_unicode_string_type

if TYPE_CHECKING:
    import numpy.typing as npt

LiteralValue = TypeVar("LiteralValue")
LiteralType = tuple[str, str, LiteralValue]
ConjunctionType = list[LiteralType]
# Optional is part of the actual type since predicate=None
# is a sential for: All values
PredicatesType = list[ConjunctionType] | None


class DataFrameSerializer:
    """Abstract class that supports serializing DataFrames to/from minimalkv
    stores.

    :meta public:
    """

    _serializers: dict[str, "DataFrameSerializer"] = {}
    type_stable = False

    def __ne__(self, other):
        return not (self == other)

    @classmethod
    def register_serializer(cls, suffix, serializer):
        cls._serializers[suffix] = serializer

    @classmethod
    def restore_dataframe(
        cls,
        store: KeyValueStore,
        key: str,
        filter_query: str | None = None,
        columns: Iterable[str] | None = None,
        predicate_pushdown_to_io: bool = True,
        categories: Iterable[str] | None = None,
        predicates: PredicatesType | None = None,
        date_as_object: bool = False,
    ) -> pd.DataFrame:
        """Load a DataFrame from the specified store. The key is also used to
        detect the used format.

        Parameters
        ----------
        store
            store engine
        key
            Key that specifies a path where object should be
            retrieved from the store resource.
        filter_query
            Optional query to filter the DataFrame. Must adhere to the specification
            of pandas.DataFrame.query.
        columns
            Only read in listed columns. When set to None, the full file
            will be read in.
        predicate_pushdown_to_io
            Push predicates through to the I/O layer, default True. Disable
            this if you see problems with predicate pushdown for the given
            file even if the file format supports it. Note that this option
            only hides problems in the store layer that need to be addressed
            there.
        categories
            Columns that should be loaded as categoricals.
        predicates
            Optional list of predicates, like [[('x', '>', 0), ...], that are used
            to filter the resulting DataFrame, possibly using predicate pushdown,
            if supported by the file format.
            This parameter is not compatible with filter_query.

            Predicates are expressed in disjunctive normal form (DNF). This means
            that the innermost tuple describe a single column predicate. These
            inner predicate make are all combined with a conjunction (AND) into a
            larger predicate. The most outer list then combines all predicates
            with a disjunction (OR). By this, we should be able to express all
            kinds of predicates that are possible using boolean logic.
        date_as_object
            Retrieve all date columns as an object column holding datetime.date objects
            instead of pd.Timestamp. Note that this option only works for type-stable
            serializers, e.g. ``ParquetSerializer``.
        """
        if filter_query and predicates:
            raise ValueError("Can only specify one of filter_query and predicates")

        for suffix, serializer in cls._serializers.items():
            if key.endswith(suffix):
                df = serializer.restore_dataframe(
                    store,
                    key,
                    filter_query,
                    columns,
                    predicate_pushdown_to_io=predicate_pushdown_to_io,
                    categories=categories,
                    predicates=predicates,
                    date_as_object=date_as_object,
                )
                df.columns = df.columns.map(ensure_unicode_string_type)
                return df

        # No serialiser matched
        raise ValueError(f"The specified file format for '{key}' is not supported")

    def store(self, store: KeyValueStore, key_prefix: str, df: pd.DataFrame) -> str:
        """Persist a DataFrame to the specified store.

        The used store format (e.g. Parquet) will be appended to the key.

        Parameters
        ----------
        store: minimalkv.KeyValueStore
                store engine
        key_prefix: str
                Key prefix that specifies a path where object should be
                stored on the store resource. The used file format will be
                appended to the key.
        df: pandas.DataFrame or pyarrow.Table
                DataFrame that shall be persisted

        Returns
        -------
        str
            The actual key where the DataFrame is stored.
        """
        raise NotImplementedError("Abstract method called.")


def filter_df(df, filter_query=None):
    """General implementation of query filtering.

    Serialisation formats such as Parquet that support predicate push-
    down may pre-filter in their own implementations.
    """
    if df.shape[0] > 0 and filter_query is not None:
        df = df.query(filter_query)
    return df


def check_predicates(predicates: PredicatesType) -> None:
    """Check if predicates are well-formed."""
    if predicates is None:
        return

    if len(predicates) == 0:
        raise ValueError("Empty predicates")

    for conjunction_idx, conjunction in enumerate(predicates):
        if not isinstance(conjunction, list):
            raise ValueError(
                f"Invalid predicates: Conjunction {conjunction_idx} should be a "
                f"list of 3-tuples, got object of type {type(conjunction)} instead."
            )
        if len(conjunction) == 0:
            raise ValueError(
                f"Invalid predicates: Conjunction {conjunction_idx} is empty"
            )
        for clause_idx, clause in enumerate(conjunction):
            if not isinstance(clause, tuple) and len(clause) == 3:
                raise ValueError(
                    f"Invalid predicates: Clause {clause_idx} in conjunction {conjunction_idx} "
                    f"should be a 3-tuple, got object of type {type(clause)} instead"
                )
            _, op, val = clause
            if (
                isinstance(val, list)
                and any(_check_contains_null(v) for v in val)
                or _check_contains_null(val)
            ):
                raise NotImplementedError(
                    "Null-terminated binary strings are not supported as predicate values."
                )
            if (
                pd.api.types.is_scalar(val)
                and pd.isnull(val)
                and op not in ["==", "!=", "is distinct from"]
            ):
                raise ValueError(
                    f"Invalid predicates: Clause {clause_idx} in conjunction {conjunction_idx} "
                    f"with null value and operator {op}. Only operators supporting null values "
                    "are '==', '!=', 'in' and 'is distinct from'."
                )


def filter_predicates_by_column(
    predicates: PredicatesType, columns: list[str]
) -> PredicatesType | None:
    """Takes a predicate list and removes all literals which are not
    referencing one of the given column.

    .. ipython:: python

        from plateau.serialization import filter_predicates_by_column

        predicates = [[("A", "==", 1), ("B", "<", 5)], [("C", "==", 4)]]

        filter_predicates_by_column(predicates, ["A"])

    Parameters
    ----------
    predicates:
        A list of predicates to be filtered
    columns:
        A list of all columns allowed in the output
    """
    if predicates is None:
        return None
    check_predicates(predicates)
    filtered_predicates = []
    for predicate in predicates:
        new_conjunction = []
        for col, op, val in predicate:
            if col in columns:
                new_conjunction.append((col, op, val))
        if new_conjunction:
            filtered_predicates.append(new_conjunction)
    if filtered_predicates:
        return filtered_predicates
    else:
        return None


def columns_in_predicates(predicates: PredicatesType) -> set[str]:
    """Determine all columns which are mentioned in the list of predicates.

    Parameters
    ----------
    predicates:
        The predicates to be scanned.
    """
    if predicates is None:
        return set()
    check_predicates(predicates)
    # Determine the set of columns that are part of a predicate
    columns = set()
    for predicates_inner in predicates:
        for col, _, _ in predicates_inner:
            columns.add(col)
    return columns


def filter_df_from_predicates(
    df: pd.DataFrame,
    predicates: PredicatesType | None,
    strict_date_types: bool = False,
) -> pd.DataFrame:
    """Filter a `pandas.DataFrame` based on predicates in disjunctive normal
    form.

    Parameters
    ----------
    df
        The pandas DataFrame to be filtered
    predicates
        Predicates in disjunctive normal form (DNF). For a thorough documentation, see
        :class:`DataFrameSerializer.restore_dataframe`
        If None, the df is returned unmodified
    strict_date_types
        If False (default), cast all datelike values to datetime64 for comparison.

    See Also
    --------
    * :ref:`predicate_pushdown`
    """
    if predicates is None:
        return df
    indexer: npt.NDArray[np.bool_] = np.zeros(len(df), dtype=bool)
    for conjunction in predicates:
        inner_indexer: npt.NDArray[np.bool_] = np.ones(len(df), dtype=bool)
        for column, op, value in conjunction:
            column_name = ensure_unicode_string_type(column)
            filter_array_like(
                df[column_name].values,
                op,
                value,
                inner_indexer,
                inner_indexer,
                strict_date_types=strict_date_types,
                column_name=column_name,
            )
        indexer = inner_indexer | indexer
    return df[indexer]


def _handle_categorical_data(array_like, require_ordered):
    if require_ordered and isinstance(array_like.dtype, pd.CategoricalDtype):
        if isinstance(array_like, pd.Categorical):
            categorical = array_like
        else:
            categorical = array_like.cat
        array_value_type = categorical.categories.dtype
        if categorical.categories.is_monotonic_increasing:
            array_like = categorical.as_ordered()
        else:
            array_like = categorical.reorder_categories(
                categorical.categories.sort_values(), ordered=True
            )
    else:
        array_value_type = array_like.dtype
    return array_like, array_value_type


def _handle_null_arrays(array_like, value_dtype):
    # NULL types might not be preserved well, so try to cast floats (pandas default type) to the value type
    # Determine the type using the `kind` interface since this is common for a numpy array, pandas series and pandas extension arrays
    if array_like.dtype.kind == "f" and np.isnan(array_like).all():
        if array_like.dtype.kind != value_dtype.kind:
            if isinstance(value_dtype, pd.api.extensions.ExtensionDtype):
                array_like = pd.Series(array_like, dtype=value_dtype).values
            else:
                array_like = array_like.astype(value_dtype)
    return array_like, array_like.dtype


def _handle_timelike_values(array_value_type, value, value_dtype, strict_date_types):
    def _cast_one_value(val):
        return pd.Series(val, dtype=array_value_type).iloc[0].to_datetime64()

    if is_list_like(value):
        value = [_cast_one_value(val) for val in value]
    else:
        value = _cast_one_value(value)
    value_dtype = pd.Series(value).dtype
    return value, value_dtype


def _ensure_type_stability(
    array_like, value, strict_date_types, require_ordered, column_name=None
):
    """Ensure that the provided value and the provided array will have
    compatible types, such that comparisons are unambiguous.

    The type check is based on the numpy type system and accesses the arrays
    `kind` attribute and asserts equality. The provided value will be
    interpreted as a scalar in this case. For scalars which do not have a proper
    python representation, we will relax the strictness as long as there is a
    valid and unambiguous interpretation of a comparison operation. In
    particular we consider the following combinations valid:

        * unsigned integer (u) <> integer (i)
        * zero-terminated bytes (S) <> Python Object (O)
        * Unicode string (U) <> Python Object (O)

    Parameters
    ----------
    strict_date_types: bool
        If False, assume that datetime.date and datetime.datetime are
        compatible types. In this case, the value is cast appropriately
    require_ordered: bool
        Indicate if the operator to be evaluated will require a notion of
        ordering. In the case of pd.Categorical we will then assume a
        lexicographical ordering and cast the pd.CategoricalDtype accordingly
    column_name: str, optional
        Name of the column where `array_like` originates from, used for nicer
        error messages.
    """

    value_dtype = pd.Series(value if is_list_like(value) else [value]).dtype
    array_like, array_value_type = _handle_categorical_data(array_like, require_ordered)
    array_like, array_value_type = _handle_null_arrays(array_like, value_dtype)

    compatible_types = [
        # UINT and INT
        ("u", "i"),
        ("i", "u"),
        # various string kinds
        ("O", "S"),
        ("O", "U"),
        ("S", "O"),
        ("U", "O"),
        # bool w/ Nones
        ("b", "O"),
    ]

    if not strict_date_types:
        # objects (datetime.date) and datetime64
        compatible_types.append(("O", "M"))

    type_comp = (value_dtype.kind, array_value_type.kind)

    if len(set(type_comp)) > 1 and type_comp not in compatible_types:
        if column_name is None:
            column_name = "<unknown>"
        raise TypeError(
            f"Unexpected type for predicate: Column {column_name!r} has pandas "
            f"type '{array_value_type}', but predicate value {value!r} has "
            f"pandas type '{value_dtype}' (Python type '{type(value)}')."
        )
    if "M" in type_comp:
        value, value_dtype = _handle_timelike_values(
            array_value_type, value, value_dtype, strict_date_types
        )
    return array_like, value


def filter_array_like(
    array_like,
    op: str,
    value,
    mask=None,
    out=None,
    strict_date_types: bool = False,
    column_name: str | None = None,
):
    """Filter an array-like object using operations defined in the predicates.

    Parameters
    ----------
    array_like
        The array like object to be filtered

        See also `pandas.api.types.is_array_like`
    op
    value
    mask
        A boolean array like object which will be combined with the result
        of this evaluation using a logical AND. If an array with all True is
        given, it will be the same result as if left empty
    out
        An array into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated
        array is returned.
    strict_date_types
        If False (default), cast all datelike values to datetime64 for comparison.
    column_name
        Name of the column where `array_like` originates from, used for nicer
        error messages.

    See Also
    --------
    * :ref:`predicate_pushdown`
    """
    if mask is None:
        mask = np.ones(len(array_like), dtype=bool)

    if out is None:
        out = np.zeros(len(array_like), dtype=bool)

    # In the case of an empty list, don't bother with evaluating types, etc.
    if is_list_like(value) and len(value) == 0:
        false_arr: npt.NDArray[np.bool_] = np.zeros(len(array_like), dtype=bool)
        np.logical_and(false_arr, mask, out=out)
        return out

    require_ordered = "<" in op or ">" in op
    array_like, value = _ensure_type_stability(
        array_like, value, strict_date_types, require_ordered, column_name
    )

    with np.errstate(invalid="ignore"):
        if op == "==":
            # If the RHS is null, carry out a null check instead of equality.
            if pd.isnull(value):
                np.logical_and(pd.isnull(array_like), mask, out=out)
            else:
                np.logical_and(array_like == value, mask, out=out)
        elif op == "!=":
            warnings.warn(
                "The current behaviour of the != will be deprecated in version 5.0.0. "
                "The new behaviour will be SQL-like, i.e. NaN values will be filtered out.",
                DeprecationWarning,
                stacklevel=1,
            )
            if pd.isnull(value):
                np.logical_and(~pd.isnull(array_like), mask, out=out)
            else:
                np.logical_and(array_like != value, mask, out=out)
        elif op == "is distinct from":
            # Currently the same as != which will have a change in behaviour in 5.0.
            if pd.isnull(value):
                np.logical_and(~pd.isnull(array_like), mask, out=out)
            else:
                np.logical_and(array_like != value, mask, out=out)
        elif op == "<=":
            np.logical_and(array_like <= value, mask, out=out)
        elif op == ">=":
            np.logical_and(array_like >= value, mask, out=out)
        elif op == "<":
            np.logical_and(array_like < value, mask, out=out)
        elif op == ">":
            np.logical_and(array_like > value, mask, out=out)
        elif op == "in":
            value = np.asarray(value)
            nullmask = pd.isnull(value)
            if value.dtype.kind in ("U", "S", "O"):
                # See GH358

                # If the values include duplicates, this would blow up with the
                # join below, rendering the mask useless
                unique_vals = np.unique(value[~nullmask])
                value_ser = pd.Series(unique_vals, name="value")
                if value_ser.dtype.kind == "S":
                    # pandas 1.4.3+ supports S|*-typed series but cannot handle null
                    # values inside of them. Thus fall back to the old behaviour of
                    # using an object typed series.
                    value_ser = value_ser.astype("object")
                arr_ser = pd.Series(array_like, name="array").to_frame()
                matching_idx = (
                    ~arr_ser.merge(
                        value_ser, left_on="array", right_on="value", how="left"
                    )
                    .value.isna()
                    .values
                )
            else:
                matching_idx = (
                    np.isin(array_like, value)
                    if len(value) > 0
                    else np.zeros(len(array_like), dtype=bool)
                )

            if any(nullmask):
                matching_idx |= pd.isnull(array_like)

            np.logical_and(
                matching_idx,
                mask,
                out=out,
            )
        else:
            raise NotImplementedError("op not supported")

    return out



import pickle

import dask
import dask.dataframe as dd
import pandas as pd
import pytest

from plateau.io.dask.dataframe import update_dataset_from_ddf
from plateau.io.testing.update import *  # noqa


@pytest.fixture
def bound_update_dataset():
    return _update_dataset


def _id(part):
    if isinstance(part, pd.DataFrame):
        return part
    else:
        return part[0]


def _update_dataset(partitions, *args, **kwargs):
    # TODO: Simplify once parse_input_to_metapartition is removed / obsolete

    with dask.config.set({"dataframe.convert-string": False}):
        if isinstance(partitions, pd.DataFrame):
            partitions = dd.from_pandas(partitions, npartitions=1)
        elif partitions is not None:
            partitions = dd.from_map(_id, partitions)
        else:
            partitions = None

    # Replace `table_name` with `table` keyword argument to enable shared test code
    # via `bound_update_dataset` fixture
    if "table_name" in kwargs:
        kwargs["table"] = kwargs["table_name"]
        del kwargs["table_name"]

    ddf = update_dataset_from_ddf(partitions, *args, **kwargs)

    s = pickle.dumps(ddf, pickle.HIGHEST_PROTOCOL)
    ddf = pickle.loads(s)

    return ddf.compute()


def _return_none():
    return None


@pytest.mark.parametrize("shuffle", [True, False])
def test_update_dataset_from_ddf_empty(store_factory, shuffle):
    with pytest.raises(ValueError) as exc_info:
        update_dataset_from_ddf(
            dd.from_map(lambda x: x, [], meta=(("a", int),)),
            store_factory,
            dataset_uuid="output_dataset_uuid",
            table="core",
            shuffle=shuffle,
            partition_on=["a"],
        ).compute()
    assert (
        str(exc_info.value)
        in [
            "Cannot store empty datasets, partition_list must not be empty if in store mode.",  # dask > 2021.5.0 + shuffle == True
            "No data left to save outside partition columns",  # dask > 2021.5.0 + shuffle == False
            "All `iterables` must have a non-zero length",  # dask >= 2024.3.0 + dask-expr used
        ]
    )



from minimalkv import get_store_from_url

from plateau.serialization import ParquetSerializer
from plateau.serialization.testing import get_dataframe_not_nested


class TimeRestore:
    """An example benchmark that times the performance of various kinds of
    iterating over dictionaries in Python."""

    params = [(10**3, 10**4), (10, 10**2, 10**3)]
    param_names = ["num_rows", "chunk_size"]

    def setup(self, num_rows, chunk_size):
        self.df = get_dataframe_not_nested(num_rows)
        self.serialiser = ParquetSerializer(chunk_size=chunk_size)
        self.store = get_store_from_url("memory://")
        self.key = self.serialiser.store(self.store, "key_prefix", self.df)
        self.predicates = [[("int16", "==", 123)]]

    def time_predicate_pushdown(self, num_rows, chunk_size):
        self.serialiser.restore_dataframe(
            self.store,
            self.key,
            predicate_pushdown_to_io=True,
            predicates=self.predicates,
        )



from plateau.core._zmsgpack import packb, unpackb


def test_msgpack():
    dct = {"a": 1, "b": {"c": "ÖaŒ"}}
    assert dct == unpackb(packb(dct))


def test_msgpack_storage(store):
    dct = {"a": 1, "b": {"c": "ÖaŒ"}}
    key = "test"
    store.put(key, packb(dct))
    value = store.get(key)
    assert dct == unpackb(value)



from collections.abc import Sequence
from functools import partial
from typing import cast

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd

from plateau.core._compat import PANDAS_3
from plateau.core.typing import StoreFactory
from plateau.io.dask.compression import pack_payload, unpack_payload_pandas
from plateau.io_components.metapartition import (
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.write import write_partition
from plateau.serialization import DataFrameSerializer

_KTK_HASH_BUCKET = "__KTK_HASH_BUCKET"


def _hash_bucket(df: pd.DataFrame, subset: Sequence[str] | None, num_buckets: int):
    """Categorize each row of `df` based on the data in the columns `subset`
    into `num_buckets` values.

    This is based on `pandas.util.hash_pandas_object`
    """

    if not subset:
        subset = df.columns
    hash_arr = pd.util.hash_pandas_object(df[subset], index=False)
    buckets = hash_arr % num_buckets

    available_bit_widths = np.array([8, 16, 32, 64])
    mask = available_bit_widths > np.log2(num_buckets)
    bit_width = min(available_bit_widths[mask])
    return df.assign(**{_KTK_HASH_BUCKET: buckets.astype(f"uint{bit_width}")})


def shuffle_store_dask_partitions(
    ddf: dd.DataFrame,
    table: str,
    secondary_indices: list[str],
    metadata_version: int,
    partition_on: list[str],
    store_factory: StoreFactory,
    df_serializer: DataFrameSerializer | None,
    dataset_uuid: str,
    num_buckets: int,
    sort_partitions_by: list[str],
    bucket_by: Sequence[str],
) -> da.Array:
    """Perform a dataset update with dask reshuffling to control partitioning.

    The shuffle operation will perform the following steps

    1. Pack payload data

       Payload data is serialized and compressed into a single byte value using
       ``distributed.protocol.serialize_bytes``, see also ``pack_payload``.

    2. Apply bucketing

       Hash the column subset ``bucket_by`` and distribute the hashes in
       ``num_buckets`` bins/buckets. Internally every bucket is identified by an
       integer and we will create one physical file for every bucket ID. The
       bucket ID is not exposed to the user and is dropped after the shuffle,
       before the store. This is done since we do not want to guarantee at the
       moment, that the hash function remains stable.

    3. Perform shuffle (dask.DataFrame.groupby.apply)

        The groupby key will be the combination of ``partition_on`` fields and the
        hash bucket ID. This will create a physical file for every unique tuple
        in ``partition_on + bucket_ID``. The function which is applied to the
        dataframe will perform all necessary subtask for storage of the dataset
        (partition_on, index calc, etc.).

    4. Unpack data (within the apply-function)

        After the shuffle, the first step is to unpack the payload data since
        the follow up tasks will require the full dataframe.

    5. Pre storage processing and parquet serialization

        We apply important pre storage processing like sorting data, applying
        final partitioning (at this time there should be only one group in the
        payload data but using the ``MetaPartition.partition_on`` guarantees the
        appropriate data structures plateau expects are created.).
        After the preprocessing is done, the data is serialized and stored as
        parquet. The applied function will return an (empty) MetaPartition with
        indices and metadata which will then be used to commit the dataset.

    Returns
    -------

    A dask.Array holding relevant MetaPartition objects as values
    """
    if ddf.npartitions == 0:
        return ddf

    group_cols = partition_on.copy()

    if num_buckets is None:
        raise ValueError("``num_buckets`` must not be None when shuffling data.")

    meta = ddf._meta
    meta[_KTK_HASH_BUCKET] = np.uint64(0)
    ddf = ddf.map_partitions(_hash_bucket, bucket_by, num_buckets, meta=meta)
    group_cols.append(_KTK_HASH_BUCKET)

    unpacked_meta = ddf._meta

    ddf2 = pack_payload(ddf, group_key=group_cols)
    if PANDAS_3:
        ddf_grouped = ddf2.shuffle(on=group_cols)

        unpack = partial(
            _unpack_store_partition,
            secondary_indices=secondary_indices,
            sort_partitions_by=sort_partitions_by,
            table=table,
            dataset_uuid=dataset_uuid,
            partition_on=partition_on,
            store_factory=store_factory,
            df_serializer=df_serializer,
            metadata_version=metadata_version,
            unpacked_meta=unpacked_meta,
        )
        return cast(
            da.Array,  # Output type depends on meta but mypy cannot infer this easily.
            ddf_grouped.map_partitions(unpack, meta=("MetaPartition", "object")),
        )
    else:
        ddf_grouped = ddf2.groupby(by=group_cols)

        unpack = partial(
            _unpack_store_partition,
            secondary_indices=secondary_indices,
            sort_partitions_by=sort_partitions_by,
            table=table,
            dataset_uuid=dataset_uuid,
            partition_on=partition_on,
            store_factory=store_factory,
            df_serializer=df_serializer,
            metadata_version=metadata_version,
            unpacked_meta=unpacked_meta,
        )
        return cast(
            da.Array,  # Output type depends on meta but mypy cannot infer this easily.
            ddf_grouped.apply(unpack, meta=("MetaPartition", "object")),
        )


def _unpack_store_partition(
    df: pd.DataFrame,
    secondary_indices: list[str],
    sort_partitions_by: list[str],
    table: str,
    dataset_uuid: str,
    partition_on: list[str],
    store_factory: StoreFactory,
    df_serializer: DataFrameSerializer | None,
    metadata_version: int,
    unpacked_meta: pd.DataFrame,
) -> MetaPartition:
    """Unpack payload data and store partition."""
    df2 = unpack_payload_pandas(df, unpacked_meta)
    kwargs = dict(
        secondary_indices=secondary_indices,
        sort_partitions_by=sort_partitions_by,
        dataset_table_name=table,
        dataset_uuid=dataset_uuid,
        partition_on=partition_on,
        store_factory=store_factory,
        df_serializer=df_serializer,
        metadata_version=metadata_version,
    )
    if PANDAS_3 and _KTK_HASH_BUCKET in df2:
        mps = df2.groupby(
            _KTK_HASH_BUCKET,
            observed=True,
        ).apply(write_partition, **kwargs)
        if mps.empty:
            return MetaPartition(None)
        if isinstance(mps, pd.Series):
            return parse_input_to_metapartition(mps.to_list())
        else:
            return parse_input_to_metapartition(mps[mps.columns[-1]].to_list())
    else:
        return write_partition(partition_df=df2, **kwargs)



# pylint: disable=E1101


import pandas as pd
import pytest
from packaging import version

from plateau.core.dataset import DatasetMetadata
from plateau.core.index import ExplicitSecondaryIndex
from plateau.core.testing import TIME_TO_FREEZE_ISO
from plateau.io_components.metapartition import MetaPartition
from plateau.io_components.read import dispatch_metapartitions
from plateau.io_components.write import (
    raise_if_dataset_exists,
    store_dataset_from_partitions,
)


def test_store_dataset_from_partitions(meta_partitions_files_only, store, frozen_time):
    dataset = store_dataset_from_partitions(
        partition_list=meta_partitions_files_only,
        dataset_uuid="dataset_uuid",
        store=store,
        dataset_metadata={"some": "metadata"},
    )

    expected_metadata = {"some": "metadata", "creation_time": TIME_TO_FREEZE_ISO}

    assert dataset.metadata == expected_metadata
    assert sorted(dataset.partitions.values(), key=lambda x: x.label) == sorted(
        (mp.partition for mp in meta_partitions_files_only), key=lambda x: x.label
    )
    assert dataset.uuid == "dataset_uuid"

    store_files = list(store.keys())
    # Dataset metadata: 1 file
    expected_number_files = 1
    # common metadata for v4 datasets
    expected_number_files += 1
    assert len(store_files) == expected_number_files

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    assert dataset == stored_dataset


def test_store_dataset_from_partitions_update(store, metadata_version, frozen_time):
    mp1 = MetaPartition(
        label="cluster_1",
        data=pd.DataFrame({"p": [1]}),
        file="1.parquet",
        indices={"p": ExplicitSecondaryIndex("p", index_dct={1: ["cluster_1"]})},
        metadata_version=metadata_version,
    )
    mp2 = MetaPartition(
        label="cluster_2",
        data=pd.DataFrame({"p": [2]}),
        file="2.parquet",
        indices={"p": ExplicitSecondaryIndex("p", index_dct={2: ["cluster_2"]})},
        metadata_version=metadata_version,
    )
    dataset = store_dataset_from_partitions(
        partition_list=[mp1, mp2],
        dataset_uuid="dataset_uuid",
        store=store,
        dataset_metadata={"dataset": "metadata"},
    )
    dataset = dataset.load_index("p", store)

    mp3 = MetaPartition(
        label="cluster_3",
        data=pd.DataFrame({"p": [3]}),
        file="3.parquet",
        indices={"p": ExplicitSecondaryIndex("p", index_dct={3: ["cluster_3"]})},
        metadata_version=metadata_version,
    )

    dataset_updated = store_dataset_from_partitions(
        partition_list=[mp3],
        dataset_uuid="dataset_uuid",
        store=store,
        dataset_metadata={"extra": "metadata"},
        update_dataset=dataset,
        remove_partitions=["cluster_1"],
    )
    dataset_updated = dataset_updated.load_index("p", store)
    expected_metadata = {"dataset": "metadata", "extra": "metadata"}

    expected_metadata["creation_time"] = TIME_TO_FREEZE_ISO

    assert dataset_updated.metadata == expected_metadata
    assert list(dataset.partitions) == ["cluster_1", "cluster_2"]
    assert list(dataset_updated.partitions) == ["cluster_2", "cluster_3"]
    assert dataset_updated.partitions["cluster_3"] == mp3.partition
    assert dataset_updated.uuid == "dataset_uuid"

    store_files = list(store.keys())
    # 1 dataset metadata file and 1 index file
    # note: the update writes a new index file but due to frozen_time this gets
    # the same name as the previous one and overwrites it.
    expected_number_files = 2
    # common metadata for v4 datasets (1 table)
    expected_number_files += 1
    assert len(store_files) == expected_number_files

    assert dataset.indices["p"].index_dct == {1: ["cluster_1"], 2: ["cluster_2"]}
    assert dataset_updated.indices["p"].index_dct == {
        2: ["cluster_2"],
        3: ["cluster_3"],
    }

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    stored_dataset = stored_dataset.load_index("p", store)
    assert dataset_updated == stored_dataset


def test_raise_if_dataset_exists(store_factory, dataset_function):
    raise_if_dataset_exists(dataset_uuid="ThisDoesNotExist", store=store_factory)
    with pytest.raises(RuntimeError):
        raise_if_dataset_exists(dataset_uuid=dataset_function.uuid, store=store_factory)


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("2"),
    reason="Timestamp unit coercion is only relevant in pandas >= 2",
)
def test_coerce_schema_timestamp_units(store):
    date = pd.Timestamp(2000, 1, 1)

    mps_original = [
        MetaPartition(label="one", data=pd.DataFrame({"a": date, "b": [date]})),
        MetaPartition(
            label="two",
            data=pd.DataFrame({"a": date.as_unit("ns"), "b": [date.as_unit("ns")]}),
        ),
    ]

    mps = (
        mp.store_dataframes(store, dataset_uuid="dataset_uuid") for mp in mps_original
    )

    # Expect this not to fail even though the metapartitions have different
    # timestamp units, because all units should be coerced to nanoseconds.
    dataset = store_dataset_from_partitions(
        partition_list=mps,
        dataset_uuid="dataset_uuid",
        store=store,
        dataset_metadata={"some": "metadata"},
    )

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    assert dataset == stored_dataset

    mps = dispatch_metapartitions("dataset_uuid", store)
    mps_loaded = (mp.load_dataframes(store) for mp in mps)

    # Ensure the values and dtypes of the loaded datasets are correct
    for mp in mps_loaded:
        assert mp.data["a"].dtype == "datetime64[ns]"
        assert mp.data["b"].dtype == "datetime64[ns]"
        assert mp.data["a"][0] == date
        assert mp.data["b"][0] == date



import io

import pytest

from plateau.serialization._io_buffer import BlockBuffer


class _ReadRecordWrapper:
    """Wrapper around file-like objects to record read requests."""

    def __init__(self, raw):
        self.raw = raw
        self.records = []

    def __getattr__(self, attr):
        return getattr(self.raw, attr)

    def read(self, size):
        pos = self.raw.tell()
        self.records.append((pos, size))
        return self.raw.read(size)


class _ZeroFile:
    """Simulated files filled with NULL-bytes, can be used to test behavior on
    very large files."""

    def __init__(self, size):
        self._size = size
        self._pos = 0

    def tell(self):
        return self._pos

    def seek(self, offset, whence=0):
        if whence == 0:
            self._pos = max(0, min(offset, self._size))
        elif whence == 1:
            self._pos = max(0, min(self._pos + offset, self._size))
        elif whence == 2:
            self._pos = max(0, min(self._size + offset, self._size))
        else:
            raise ValueError("unsupported whence value")

        return self._pos

    def read(self, size=None):
        if (size is None) or (size < 0) or (self._pos + size > self._size):
            # read entire, remaining file
            size = self._size - self._pos

        return b"\0" * size

    def seekable(self):
        return True

    def readable(self):
        return True

    @property
    def closed(self):
        return False


@pytest.fixture
def raw_inner():
    return io.BytesIO(b"foxbar")


@pytest.fixture
def raw(raw_inner):
    return _ReadRecordWrapper(raw_inner)


@pytest.fixture(params=[1, 3, 4, 6, 7, 10])
def blocksize(raw, request):
    return request.param


@pytest.fixture
def example_buffer(raw, blocksize):
    return BlockBuffer(raw, blocksize)


def test_init_fails_not_seekable():
    raw = io.BytesIO()
    raw.seekable = lambda: False
    with pytest.raises(ValueError, match="raw must be seekable"):
        BlockBuffer(raw)


def test_init_fails_not_readable():
    raw = io.BytesIO()
    raw.readable = lambda: False
    with pytest.raises(ValueError, match="raw must be readable"):
        BlockBuffer(raw)


@pytest.mark.parametrize("blocksize", [-1, 0])
def test_init_fails_blocksize(blocksize):
    raw = io.BytesIO()
    with pytest.raises(ValueError, match="blocksize must be at least 1"):
        BlockBuffer(raw, blocksize)


def test_init_fails_closed(blocksize):
    raw = io.BytesIO()
    raw.close()
    with pytest.raises(ValueError, match="Cannot use closed file object"):
        BlockBuffer(raw, blocksize)


def test_size(example_buffer):
    assert example_buffer.size == 6


def test_seekable(example_buffer):
    assert example_buffer.seekable()


def test_readable(example_buffer):
    assert example_buffer.readable()


@pytest.mark.parametrize("size", [None, -1, -100])
def test_read_all(example_buffer, size):
    assert example_buffer.read(size) == b"foxbar"
    assert example_buffer.tell() == 6


def test_read_sequential(example_buffer):
    assert example_buffer.read(2) == b"fo"
    assert example_buffer.tell() == 2
    assert example_buffer.read(3) == b"xba"
    assert example_buffer.tell() == 5
    assert example_buffer.read(0) == b""
    assert example_buffer.tell() == 5
    assert example_buffer.read(2) == b"r"
    assert example_buffer.tell() == 6
    assert example_buffer.read(2) == b""
    assert example_buffer.tell() == 6


def test_fail_illegal_whence(example_buffer):
    with pytest.raises(ValueError, match="unsupported whence value"):
        example_buffer.seek(0, -1)


def test_seek_absolute_and_read(example_buffer):
    assert example_buffer.seek(2) == 2
    assert example_buffer.tell() == 2
    assert example_buffer.read(3) == b"xba"

    assert example_buffer.seek(2) == 2
    assert example_buffer.tell() == 2
    assert example_buffer.read(3) == b"xba"

    # beyond end
    assert example_buffer.seek(100) == 6
    assert example_buffer.tell() == 6
    assert example_buffer.read(3) == b""

    # beyond begin
    assert example_buffer.seek(-1) == 0
    assert example_buffer.tell() == 0
    assert example_buffer.read(1) == b"f"


def test_seek_relative_and_read(example_buffer):
    assert example_buffer.seek(2, 1) == 2
    assert example_buffer.tell() == 2
    assert example_buffer.read(2) == b"xb"

    assert example_buffer.tell() == 4
    assert example_buffer.seek(1, 1) == 5
    assert example_buffer.tell() == 5
    assert example_buffer.read(1) == b"r"

    assert example_buffer.tell() == 6
    assert example_buffer.seek(-3, 1) == 3
    assert example_buffer.tell() == 3
    assert example_buffer.read(1) == b"b"

    assert example_buffer.tell() == 4
    assert example_buffer.seek(0, 1) == 4
    assert example_buffer.tell() == 4
    assert example_buffer.read(1) == b"a"

    # beyond end
    assert example_buffer.tell() == 5
    assert example_buffer.seek(100, 1) == 6
    assert example_buffer.tell() == 6
    assert example_buffer.read(3) == b""

    # beyond begin
    assert example_buffer.seek(-100, 1) == 0
    assert example_buffer.tell() == 0
    assert example_buffer.read(1) == b"f"


def test_seek_reverse_and_read(example_buffer):
    assert example_buffer.seek(-4, 2) == 2
    assert example_buffer.tell() == 2
    assert example_buffer.read(3) == b"xba"

    assert example_buffer.seek(-4, 2) == 2
    assert example_buffer.tell() == 2
    assert example_buffer.read(3) == b"xba"

    # beyond end
    assert example_buffer.seek(1, 2) == 6
    assert example_buffer.tell() == 6
    assert example_buffer.read(3) == b""

    # beyond begin
    assert example_buffer.seek(-10, 2) == 0
    assert example_buffer.tell() == 0
    assert example_buffer.read(1) == b"f"


def test_caching_read_all(raw, example_buffer):
    example_buffer.read()
    example_buffer.read()
    assert raw.records == [(0, 6)]


def test_caching_reuse(raw):
    b = BlockBuffer(raw, 3)

    b.read(1)
    assert raw.records == [(0, 3)]

    b.seek(0)
    b.read(2)
    assert raw.records == [(0, 3)]


def test_caching_remainder(raw):
    b = BlockBuffer(raw, 4)

    b.seek(5)
    b.read()
    assert raw.records == [(4, 2)]


def test_caching_chunks1(raw):
    b = BlockBuffer(raw, 2)

    b.read(1)
    assert raw.records == [(0, 2)]

    b.seek(0)
    b.read()
    assert raw.records == [(0, 2), (2, 4)]


def test_caching_chunks2(raw):
    b = BlockBuffer(raw, 2)

    b.seek(3)
    b.read(1)
    assert raw.records == [(2, 2)]

    b.seek(0)
    b.read()
    assert raw.records == [(2, 2), (0, 2), (4, 2)]


def test_empty(blocksize):
    raw_inner = io.BytesIO()
    raw = _ReadRecordWrapper(raw_inner)
    b = BlockBuffer(raw, blocksize)

    assert b.size == 0
    assert b.read() == b""
    assert raw.records == []


def test_giga():
    raw_size = 100 * 1024**3  # 100 GB
    raw_inner = _ZeroFile(raw_size)
    raw = _ReadRecordWrapper(raw_inner)
    blocksize = 4 * 1024**2  # 4MB
    b = BlockBuffer(raw, blocksize)

    assert b.size == raw_size
    assert b.read(10) == b"\0" * 10
    assert b.seek(-10, 2) == raw_size - 10
    assert b.read(10) == b"\0" * 10
    assert raw.records == [(0, blocksize), (raw_size - blocksize, blocksize)]


def test_closed(blocksize):
    raw = io.BytesIO()
    b = BlockBuffer(raw, blocksize)
    b.close()

    assert b.closed
    assert raw.closed

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.size

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.seekable()

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.readable()

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.tell()

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.seek(0)

    with pytest.raises(ValueError, match="I/O operation on closed file."):
        b.read()

    # closing twice works
    b.close()


def test_real_file(tmpdir, blocksize):
    path = tmpdir.join("test_real_file.bin").strpath
    with open(path, "wb") as fp:
        fp.write(b"foxbar")

    real_file = open(path, "rb")

    b = BlockBuffer(real_file, blocksize)

    assert not b.closed

    assert b.size == 6
    assert b.seekable() is True
    assert b.readable() is True
    assert b.tell() == 0
    assert b.seek(1) == 1
    assert b.read() == b"oxbar"

    # final close
    b.close()

    # closing twice works
    b.close()



import datetime

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
from packaging import version

from plateau.io.eager import (
    read_dataset_as_dataframes,
    read_table,
    store_dataframes_as_dataset,
)
from plateau.io.testing.read import *  # noqa

PANDAS_LT_3 = version.parse(pd.__version__) < version.parse("3.0.0.dev")


@pytest.fixture(
    params=["dataframe", "table"],
    ids=["dataframe", "table"],
)
def output_type(request):
    # TODO: get rid of this parametrization and split properly into two functions
    return request.param


def _read_table(*args, **kwargs):
    kwargs.pop("dispatch_by", None)
    res = read_table(*args, **kwargs)

    if len(res):
        # Array split conserves dtypes
        dfs = np.array_split(res, len(res))
        if PANDAS_LT_3:
            return dfs
        return [
            # Workaround for https://github.com/numpy/numpy/issues/24889.
            pd.DataFrame(data=arr, columns=res.columns).astype(res.dtypes)
            for arr in dfs
        ]
    else:
        return [res]


# FIXME: handle removal of metparittion function properly.
# FIXME: consolidate read_Dataset_as_dataframes (replaced by iter)
def _read_dataset(output_type, *args, **kwargs):
    if output_type == "table":
        return _read_table
    elif output_type == "dataframe":
        return read_dataset_as_dataframes
    else:
        raise NotImplementedError()


@pytest.fixture()
def bound_load_dataframes(output_type):
    return _read_dataset(output_type)


@pytest.fixture()
def backend_identifier():
    return "eager"


def test_read_table_eager(dataset, store_session, use_categoricals):
    if use_categoricals:
        categories = ["P"]
    else:
        categories = None

    df = read_table(
        store=store_session,
        dataset_uuid="dataset_uuid",
        categoricals=categories,
    )
    expected_df = pd.DataFrame(
        {
            "P": [1, 2],
            "L": [1, 2],
            "TARGET": [1, 2],
            "DATE": [datetime.date(2010, 1, 1), datetime.date(2009, 12, 31)],
        }
    )
    if categories:
        expected_df = expected_df.astype({"P": "category"})

    # No stability of partitions
    df = df.sort_values(by="P").reset_index(drop=True)

    pdt.assert_frame_equal(df, expected_df, check_dtype=True, check_like=True)


def test_read_table_with_columns(dataset, store_session):
    df = read_table(
        store=store_session,
        dataset_uuid="dataset_uuid",
        columns=["P", "L"],
    )

    expected_df = pd.DataFrame({"P": [1, 2], "L": [1, 2]})

    # No stability of partitions
    df = df.sort_values(by="P").reset_index(drop=True)
    expected_df = expected_df.sort_values(by="P").reset_index(drop=True)

    pdt.assert_frame_equal(df, expected_df, check_dtype=False, check_like=True)


def test_read_table_simple_list_for_cols_cats(dataset, store_session):
    df = read_table(
        store=store_session,
        dataset_uuid="dataset_uuid",
        columns=["P", "L"],
        categoricals=["P", "L"],
    )

    expected_df = pd.DataFrame({"P": [1, 2], "L": [1, 2]})

    # No stability of partitions
    df = df.sort_values(by="P").reset_index(drop=True)
    expected_df = expected_df.sort_values(by="P").reset_index(drop=True)

    expected_df = expected_df.astype("category")

    pdt.assert_frame_equal(df, expected_df, check_dtype=False, check_like=True)


def test_write_and_read_default_table_name(store_session):
    df_write = pd.DataFrame({"P": [3.14, 2.71]})
    store_dataframes_as_dataset(store_session, "test_default_table_name", [df_write])

    # assert default table name "table" is used
    df_read_as_dfs = read_dataset_as_dataframes(
        "test_default_table_name", store_session
    )
    pd.testing.assert_frame_equal(df_write, df_read_as_dfs[0])


@pytest.mark.parametrize("partition_on", [None, ["A", "B"]])
def test_read_or_predicates(store_factory, partition_on):
    # https://github.com/JDASoftwareGroup/plateau/issues/295
    dataset_uuid = "test"
    df = pd.DataFrame({"A": range(10), "B": ["A", "B"] * 5, "C": range(-10, 0)})

    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        dfs=[df],
        partition_on=partition_on,
    )

    df1 = read_table(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        predicates=[[("A", "<", 3)], [("A", ">", 5)], [("B", "==", "non-existent")]],
    )

    df2 = read_table(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        predicates=[[("A", "<", 3)], [("A", ">", 5)]],
    )
    expected = pd.DataFrame(
        data={
            "A": [0, 1, 2, 6, 7, 8, 9],
            "B": ["A", "B", "A", "A", "B", "A", "B"],
            "C": [-10, -9, -8, -4, -3, -2, -1],
        },
    )

    pd.testing.assert_frame_equal(df1, df2)
    pd.testing.assert_frame_equal(expected, df2)
